#!/usr/bin/env
__author__ = 'Gavin M. Roy'
__email__ = 'gmr@myyearbook.com'
__since__ = '03/14/2011'
__version__ = "2.0"

# Tinman direct imports for shorter access
from cache import memoize
from whitelist import whitelisted
from utils import log_method_call
from utils import import_namespaced_class

# Import utils for main use
import utils

# Base imports for running the tinman app runner
from logging import debug, error, info
from multiprocessing import Process
from optparse import OptionParser
from os import unlink
from os.path import exists, dirname, realpath
from signal import signal, SIGTERM
from sys import stderr, exit
from time import sleep
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application

# Handle for our process wide ioloop global
ioloop = None


class TinmanApplication(Application):

    def __init__(self, routes=None, **settings):

        # If we have routes, import the classes and set everything up real nice
        if routes:

            # Validate it's a list
            if not isinstance(routes, list):
                raise ValueError("Routes should be a list of tuples")

            handlers = []
            for parts in routes:

                # Return the reference to the python class at the end of the
                # namespace. eg foo.Baz, foo.bar.Baz
                try:
                    parts[1] = import_namespaced_class(parts[1])
                except ImportError as err:
                    error("Skipping %s due to import error: %s",
                          parts[1], err)
                    continue

                # Append our handle stack
                info('Appending handler: %r', parts)
                handlers.append(tuple(parts))

            # Set the app version from the version setting in this file
            if 'version' not in settings:
                settings['version'] = __version__

            # Set the base path for use inside the app since all code locations
            # are not relative
            if 'base_path' not in settings:
                settings['base_path'] = dirname(realpath(__file__))

            # If we have a static_path
            if 'static_path' in settings:
                # Replace __base_path__ with the path this is running from
                settings['static_path'] =\
                    settings['static_path'].replace('__base_path__',
                                                    settings['base_path'])

        # If we specified the UI modules module we need to import it
        if 'ui_modiles' in settings:
            try:
                # Assign the modules to the import
                settings['ui_modules'] = \
                    import_namespaced_class(settings['ui_modules'])
            except ImportError as err:
                error("Error importing UI Modules %s: %s",
                      settings['ui_modules'], err)

        # Create our Application for this process
        Application.__init__(self, handlers, **settings)


def _shutdown_signal_handler(signum, frame):
    """
    Called on SIGTERM to shutdown the sub-process
    """
    if ioloop:
        ioloop.stop()


def start(routes, settings, http_config, port):
    global ioloop

    # Setup a signal handler to cleanly shutdown this process
    signal(SIGTERM, _shutdown_signal_handler)

    # Start our application
    application = TinmanApplication(routes, **settings)

    # Start the HTTP Server
    info("Starting Tornado HTTPServer on port %i" % port)
    http_server = HTTPServer(application)
    http_server.listen(port)

    # Get a handle to the instance of IOLoop
    ioloop = IOLoop.instance()

    # Start the IOLoop
    try:
        ioloop.start()
    except KeyboardInterrupt:
        debug("Keyboard Interrupt received, shutting down child Process")


def main(*args):
    # Setup optparse
    usage = "usage: %prog -c <configfile> [options]"
    version_string = "%%prog %s" % __version__
    description = "Tornado application runner"

    # Create our parser and setup our command line options
    parser = OptionParser(usage=usage,
                          version=version_string,
                          description=description)

    parser.add_option("-c", "--config",
                      action="store",
                      dest="config",
                      help="Specify the configuration file for use")
    parser.add_option("-f", "--foreground",
                      action="store_true",
                      dest="foreground",
                      default=False,
                      help="Run interactively in console")
    parser.add_option("-r", "--route-decorator",
                      action="store_true",
                      dest="route_decorator",
                      default=False,
                      help="Utilize the route decorator instead of the Routes\
                            section of the config")

    # Parse our options and arguments
    options, args = parser.parse_args()

    # No configuration file?
    if options.config is None:
        stderr.write('\nERROR: Missing configuration file\n\n')
        parser.print_help()
        exit(1)

    # Load in the YAML config file
    config = utils.load_configuration_file(options.config)

    # Required

    if 'Application' not in config:
        raise AttributeError("Missing Application section in configuration")

    if 'HTTPServer' not in config:
        raise AttributeError("Missing HTTPServer section in configuration")

    if 'Logging' not in config:
        raise AttributeError("Missing Logging section in configuration")

    if not options.route_decorator and 'Routes' not in config:
        raise AttributeError("Missing Routes section in configuration")

    if not options.route_decorator and \
       not isinstance(config['Routes'], list):
        raise AttributeError("Error in Routes section in configuration")

    if 'base_path' in config:
        import sys
        sys.path.insert(0, config['base_path'])

    # Setup our logging
    utils.setup_logging(config['Logging'], options.foreground)

    # Setup our signal handlers
    utils.setup_signals()

    # Daemonize if we need to
    if not options.foreground:
        utils.daemonize(pidfile=config.get("pidfile", None),
                        user=config.get("user", None),
                        group=config.get("group", None))
    else:
        config['Application']['debug'] = True

    # Loop through and kick off our processes
    children = []
    for port in config['HTTPServer']['ports']:
        debug("Starting process for port %i" % port)
        # Kick off the child process
        child = Process(target=start,
                        name="tinman-%i" % port,
                        args=(config['Routes'],
                              config['Application'],
                              config['HTTPServer'],
                              port))
        children.append(child)
        child.start()
    debug("All children spawned")

    # Have a main event loop for dealing with stats
    utils.running = True
    while utils.running:
        try:
            sleep(1)
        except KeyboardInterrupt:
            info("CTRL-C pressed, shutting down.")
            break

    # Tell all the children to shutdown
    for child in children:
        debug("Sending terminate signal to %s", child.name)
        child.terminate()

    # Loop while the children are shutting down
    debug("Waiting for children to shutdown")
    while True in [child.is_alive() for child in children]:
        sleep(0.5)

    # Remove our pidfile
    if 'pidfile' in config:
        if exists(config['pidfile']):
            try:
                unlink(config['pidfile'])
            except OSError as e:
                error("Could not remove pidfile: %s", e)

    # Log that we're shutdown cleanly
    info("tinman has shutdown")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Functions used mainly in startup and shutdown of tornado applications
"""
import grp
import logging
import os
import os.path
import pwd
import signal
import sys
import yaml

from functools import wraps
from socket import gethostname
from tornado.options import enable_pretty_logging

# Callback handlers
rehash_handler = None
shutdown_handler = None

# Application state for shutdown
running = False


def log_method_call(method):
    """
    Logging decorator to send the method and arguments to logging.debug
    """
    @wraps(method)
    def debug_log(*args, **kwargs):

        if logging.getLogger('').getEffectiveLevel() == logging.DEBUG:

            # Get the class name of what was passed to us
            try:
                class_name = args[0].__class__.__name__
            except AttributeError:
                class_name = 'Unknown'
            except IndexError:
                class_name = 'Unknown'

            # Build a list of arguments to send to the logging
            log_args = list()
            for x in xrange(1, len(args)):
                log_args.append(args[x])
            if len(kwargs) > 1:
                log_args.append(kwargs)

            # If we have arguments, log them as well, otherwise just the method
            if log_args:
                logging.debug("%s.%s(%r) Called", class_name, method.__name__,
                             log_args)
            else:
                logging.debug("%s.%s() Called", class_name, method.__name__)

        # Actually execute the method
        return method(*args, **kwargs)

    # Return the debug_log function to the python stack for execution
    return debug_log


def application_name():
    """
    Returns the currently running application name
    """
    return os.path.split(sys.argv[0])[1]


def hostname():
    """
    Returns the hostname for the machine we're running on
    """
    return gethostname().split(".")[0]


def daemonize(pidfile=None, user=None, group=None):
    """
    Fork the Python app into the background and close the appropriate
    "files" to detach from console. Based off of code by Jürgen Hermann and
    http://code.activestate.com/recipes/66012/

    Parameters:

    * pidfile: Pass in a file to write the pid, defaults to
               /tmp/current_process_name-pid_number.pid
    * user: User to run as, defaults to current user
    * group: Group to run as, defaults to current group
    """

    # Flush stdout and stderr
    sys.stdout.flush()
    sys.stderr.flush()

    # Set our default uid, gid
    uid, gid = -1, -1

    # Get the user id if we have a user set
    if user:
        uid = pwd.getpwnam(user).pw_uid

    # Get the group id if we have a group set
    if group:
        gid = grp.getgrnam(group).gr_gid

    # Fork off from the process that called us
    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    # Second fork to put into daemon mode
    pid = os.fork()
    if pid > 0:
        # exit from second parent, print eventual PID before
        sys.stdout.write('%s: started - PID # %d\n' % (application_name(),
                                                       pid))

        # Setup a pidfile if we weren't passed one
        pidfile = pidfile or \
                  os.path.normpath('/tmp/%s-%i.pid' % (application_name(),
                                                       pid))

        # Write a pidfile out
        with open(pidfile, 'w') as f:
            f.write('%i\n' % pid)

            # If we have uid or gid change the uid/gid for the file
            if uid > -1 or gid > -1:
                os.fchown(f.fileno(), uid, gid)

        # Exit the parent process
        sys.exit(0)

    # Detach from parent environment
    os.chdir(os.path.normpath('/'))
    os.umask(0)
    os.setsid()

    # Redirect stdout, stderr, stdin
    si = file('/dev/null', 'r')
    so = file('/dev/null', 'a+')
    se = file('/dev/null', 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    # Set the running user
    if user and group:
        logging.info("Changing the running user:group to %s:%s", user, group)
    elif user:
        logging.info("Changing the running user to %s", user)
    elif group:
        logging.info("Changing the group to %s", group)

    # If we have a uid and it's not for the running user
    if uid > -1 and uid != os.geteuid():
            try:
                os.seteuid(uid)
                logging.debug("User changed to %s(%i)", user, uid)
            except OSError as e:
                logging.error("Could not set the user: %s", str(e))

    # if we have a gid and it's not for the current group
    if gid > -1 and gid != os.getegid():
        try:
            os.setgid(gid)
            logging.debug("Process group changed to %s(%i)", group, gid)
        except OSError as e:
            logging.error("Could not set the group: %s", str(e))

    return True


def load_configuration_file(config_file):
    """
    Load our YAML configuration file from disk or error out
    if not found or parsable
    """
    try:
        with file(config_file, 'r') as f:
            config = yaml.load(f)

    except IOError:
        sys.stderr.write('Configuration file not found "%s"\n' % config_file)
        sys.exit(1)

    except yaml.scanner.ScannerError as err:
        sys.stderr.write('Invalid configuration file "%s":\n%s\n' % \
                         (config_file, err))
        sys.exit(1)

    return config


def setup_logging(config, debug=False):
    """
    Setup the logging module to respect our configuration values.
    Expects a dictionary called config with the following parameters

    * directory:   Optional log file output directory
    * filename:    Optional filename, not needed for syslog
    * format:      Format for non-debug mode
    * level:       One of debug, error, warning, info
    * handler:     Optional handler
    * syslog:      If handler == syslog, parameters for syslog
      * address:   Syslog address
      * facility:  Syslog facility

    Passing in debug=True will disable any log output to anything but stdout
    and will set the log level to debug regardless of the config.
    """
    # Set logging levels dictionary
    logging_levels = {'debug':    logging.DEBUG,
                      'info':     logging.INFO,
                      'warning':  logging.WARNING,
                      'error':    logging.ERROR,
                      'critical': logging.CRITICAL}

    # Get the logging value from the dictionary
    logging_level = config['level']

    if debug:

        # Override the logging level to use debug mode
        config['level'] = logging.DEBUG

        # If we have specified a file, remove it so logging info goes to stdout
        if 'filename' in config:
            del config['filename']

    else:

        # Use the configuration option for logging
        config['level'] = logging_levels.get(config['level'], logging.NOTSET)

    # Pass in our logging config
    logging.basicConfig(**config)
    logging.info('Log level set to %s' % logging_level)

    # Get the default logger
    default_logging = logging.getLogger()

    # Remove the default stream handler
    stream_handler = None
    for handler in default_logging.handlers:
        if isinstance(handler, logging.StreamHandler):
            stream_handler = handler
            break

    # Use colorized output
    if config['level'] == logging.DEBUG:
        enable_pretty_logging()

    # If we have supported handler
    elif 'handler' in config:

        # If we want to syslog
        if config['handler'] == 'syslog':

            facility = config['syslog']['facility']
            import logging.handlers as handlers

            # If we didn't type in the facility name
            if facility in handlers.SysLogHandler.facility_names:

                # Create the syslog handler
                address = config['syslog']['address']
                facility = handlers.SysLogHandler.facility_names[facility]
                syslog = handlers.SysLogHandler(address=address,
                                                facility=facility)
                # Add the handler
                default_logging.addHandler(syslog)

                # Remove the StreamHandler
                if stream_handler:
                    default_logging.removeHandler(stream_handler)
            else:
                logging.error('%s:Invalid facility, syslog logging aborted',
                              application_name())


def shutdown():
    """
    Cleanly shutdown the application
    """
    global running

    # Tell all our children to stop
    if shutdown_handler:
        shutdown_handler()

    # Set the running state
    running = False


def setup_signals():
    """
    Setup the signals we want to be notified on
    """
    signal.signal(signal.SIGTERM, _shutdown_signal_handler)
    signal.signal(signal.SIGHUP, _rehash_signal_handler)


def _shutdown_signal_handler(signum, frame):
    """
    Called on SIGTERM to shutdown the application
    """
    logging.info("SIGTERM received, shutting down")
    shutdown()


def _rehash_signal_handler(signum, frame):
    """
    Would be cool to handle this and effect changes in the config
    """
    logging.info("SIGHUP received, rehashing config")
    if rehash_handler:
        rehash_handler()


def import_namespaced_class(path):
    """
    Pass in a string in the format of foo.Bar, foo.bar.Baz, foo.bar.baz.Qux
    and it will return a handle to the class
    """
    # Split up our string containing the import and class
    parts = path.split('.')

    # Build our strings for the import name and the class name
    import_name = '.'.join(parts[0:-1])
    class_name = parts[-1]

    # get the handle to the class for the given import
    class_handle = getattr(__import__(import_name, fromlist=class_name),
                           class_name)

    # Return the class handle
    return class_handle
